{+--------------------------------------------------------------------------+
 | Class:       TTFileSearch  ( TurboSearch )
 | Created:     10.98
 | Author:      Martin Waldenburg
 | Description: A very fast search engine, about twice as fast as Boyer-Moore,
 |              based on an article in the German magazine c't (8/97).
 |              However "Look_at" isn't implemented.
 |              The original is in 'C '. www.Heise.com.
 |              You can search case sensitive or case insensitive.
 | Version:     1.1
 | Copyright (c) 1998 Martin Waldenburg
 | All rights reserved.
 |
 | LICENCE CONDITIONS
 |
 | USE OF THE ENCLOSED SOFTWARE
 | INDICATES YOUR ASSENT TO THE
 | FOLLOWING LICENCE CONDITIONS.
 |
 |
 |
 | These Licence Conditions are exlusively
 | governed by the Law and Rules of the
 | Federal Republic of Germany.
 |
 | Redistribution and use in source and binary form, with or without
 | modification, are permitted provided that the following conditions
 | are met:
 |
 | 1. Software using this code must contain a visible line of credit,
 |    to Martin Waldenburg and the German magazine c't (8/97).
 |
 | DISCLAIMER:
 |
 | THIS SOFTWARE IS PROVIDED BY THE AUTHOR 'AS IS'.
 |
 | ALL EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 | PARTICULAR PURPOSE ARE DISCLAIMED.
 |
 | IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 | INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 | (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
 | OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 | INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 | WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 | NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 | THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 |
 |  Martin.Waldenburg@T-Online.de

+--------------------------------------------------------------------------+}
unit mTFiSear;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls, Forms, Dialogs;

type
  TTFileSearch=class(TObject)
  private
    fStream: TFileStream;
    fFileBuff: PChar;
    fBuffSize: LongInt;
    BuffSearchLen: LongInt;
    fBuffStart: LongInt;
    fBuffEof: Boolean;
    fFileEof: Boolean;
    Shift: array[0..255]of Integer;
    CompTable: array[#0..#255]of byte;
    Pat: String;
    fCount: LongInt;
    SimpleJump: Integer;
    fPos: LongInt;
    PatLen, PatLenPlus: Integer;
    fFinished: Boolean;
    fFound: Boolean;
    fPosition: LongInt;
    procedure MakeCompTable;
    function GetFinished: Boolean;
    function GetFileSize: LongInt;
  protected
    function FillBuff: LongInt;
  public
    Constructor Create(const FileName: string; Mode: Word; BuffLen: LongInt);
    Destructor Destroy; override;
    procedure Init(NewPattern: String);
    function Next: LongInt;
    procedure IInit(NewPattern: String);
    function INext: LongInt;
    property Count: LongInt read fCount write fCount;
    property FileSize: LongInt read GetFileSize;
    property Finished: Boolean read GetFinished;
    property Found: Boolean read fFound;
    property Position: LongInt read fPosition;
  end;

implementation

{ TTFileSearch }

procedure TTFileSearch.MakeCompTable;
var
  I: Byte;
begin
  for I:=0 to 255 do
    CompTable[Char(I)]:=ord(AnsiLowerCase(Char(I))[1]);
end; { MakeCompTable }

constructor TTFileSearch.Create(const FileName: string; Mode: Word; BuffLen: LongInt);
begin
  inherited Create;
  fStream:=TFileStream.Create(FileName, Mode);
  fFileBuff:=nil;
  if BuffLen<65533 then BuffLen:=35533;
  if fStream.Size>BuffLen then fBuffSize:=BuffLen else fBuffSize:=fStream.Size;
  GetMem(fFileBuff, fBuffSize+2);
  MakeCompTable;
end;

destructor TTFileSearch.Destroy;
begin
  FreeMem(fFileBuff, fBuffSize+2);
  fStream.Free;
  inherited Destroy;
end;

function TTFileSearch.GetFileSize: LongInt;
begin
  Result:= fStream.Size;
end;

function TTFileSearch.FillBuff: LongInt;
begin
  Result:=fStream.Read(fFileBuff^, fBuffSize);
  Case Result=fBuffSize of
    True: fFileEof:=False;
    False: fFileEof:=True;
  end;
  if fStream.Position=fStream.Size then fFileEof:=True;
  fBuffStart:=fStream.Position-Result;
end;

function TTFileSearch.GetFinished: Boolean;
begin
  fFinished:=False;
  fBuffEof:=False;
  if fPos>=BuffSearchLen then fBuffEof:=True;
  if PatLen>fStream.Size then fFileEof:=True;
  if fFileEof and fBuffEof then fFinished:=True;
  Result:=fFinished;
end;

procedure TTFileSearch.IInit(NewPattern: String);
var
  I: Byte;
begin
  Pat:=NewPattern;
  PatLen:=Length(Pat);
  PatLenPlus:=PatLen+1;
  for I:=0 to 255 do Shift[I]:=PatLenPlus;
  for I:=1 to PatLen do Shift[CompTable[Pat[i]]]:=PatLenPlus-I;
  SimpleJump := Shift[CompTable[Pat[PatLen]]];
  fStream.Position:=0;
  BuffSearchLen:=FillBuff;
  fFound:=False;
  fPosition:=0;
  fCount:=0;
  fPos:=-1;
end; { IInit }

procedure TTFileSearch.Init(NewPattern: String);
var
  I: Byte;
begin
  Pat:=NewPattern;
  PatLen:=Length(Pat);
  PatLenPlus:=PatLen+1;
  for I:=0 to 255 do Shift[I]:=PatLenPlus;
  for I:=1 to PatLen do Shift[ord(Pat[I])]:=PatLenPlus-I;
  SimpleJump := Shift[ord(Pat[PatLen])];
  fStream.Position:=0;
  BuffSearchLen:=FillBuff;
  fFound:=False;
  fPosition:=0;
  fCount:=0;
  fPos:=-1;
end; { Init }

function TTFileSearch.INext: LongInt;
var
  I, J: Integer;
  Label StartI;
begin
  StartI:;
  Result:=0;
  fFound:=False;
  inc(fPos, PatLen);
  fPosition:=0;
  while fPos<=BuffSearchLen do
  begin
    I:=PatLen;
    if(CompTable[Pat[I]]<>CompTable[fFileBuff[fPos]])then
      inc(fPos, Shift[CompTable[fFileBuff[fPos+1]]])
    else
    begin
      J:=fPos;
      repeat
        dec(I); dec(J);
      until(I=0)or(CompTable[Pat[I]]<>CompTable[fFileBuff[J]]);
      Case I=0 of
        True:
          begin
            fFound:=True;
            inc(fCount);
            fPosition:=fBuffStart+fPos-Patlen+1;
            Result:=fPosition;
            break;
          end;
        False: inc(fPos, Shift[CompTable[fFileBuff[J+1]]]);
      end;
    end;
  end;
  if(fFound = False )and(GetFinished = False)then
  begin
    fStream.Position:= fStream.Position -(PatLen -1); { could be optimized }
    BuffSearchLen:=FillBuff;
    fPosition:=0;
    fPos:=-1;
    Goto StartI;
  end;
end; { INext }

function TTFileSearch.Next: LongInt;
var
  I, J: Integer;
  Label Start;
begin
  Start:;
  Result:=0;
  fFound:=False;
  inc(fPos, PatLen);
  fPosition:=0;
  while fPos<=BuffSearchLen do
  begin
    I:=PatLen;
    if(Pat[I]<>fFileBuff[fPos])then inc(fPos, Shift[ord(fFileBuff[fPos+1])])
    else
    begin
      J:=fPos;
      repeat
        dec(I); dec(J);
      until(I=0)or(Pat[I]<>fFileBuff[J]);
      Case I=0 of
        True:
          begin
            fFound:=True;
            inc(fCount);
            fPosition:=fBuffStart+fPos-Patlen+1;
            Result:=fPosition;
            break;
          end;
        False: inc(fPos, Shift[ord(fFileBuff[J+1])]);
      end;
    end;
  end;
  if(fFound = False )and(GetFinished = False)then
  begin
    fStream.Position:= fStream.Position -(PatLen -1);  { could be optimized }
    BuffSearchLen:=FillBuff;
    fPosition:=0;
    fPos:=-1;
    GoTo Start;
  end;
end; { Next }

end.

